﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace VL.Vector.Core
{
    /// <summary>
    /// Vector is a generic type, allowing to both type its Cardinality at compile time, and its 
    /// underlying primitive (usually Float32)
    /// </summary>
    /// <typeparam name="D">One of a finely curated set of Dimensions implementation types</typeparam>
    /// <typeparam name="T">All items of the Vector will be of that type, and therefore more homogenous then say 7.0 tuples. Usually Float32, unless you want custom precision/accuracy</typeparam>
    /// <remarks>Heads up: this is currently a *class*, but can still be considered immutable. If this were a more vl conformant *struct*, 
    /// it would need additional factory infrastructure, an immutable data holder and general rethinking about initialisation.
    /// Because it is backed by an Array anyway, my guess is, that it might not be worth the effort. 
    ///
    /// Not sure about garbage collection on this one. 
    /// </remarks>
    public class Vector<T, D> : IEnumerable<T>  where D : struct, Dimensions
    {
        // internal data holder. could also be an immutable vl spread, dunno
        // array backing seems logical for fast access (within this namespace only!)
        internal readonly T[] Data;

        // immutable Dim. will be initialized depending on the generic implementation
        public readonly D Dim;

        // immutable Cardinality, Awesome! 
        public int Cardinality
        {
            get
            {
                return Dim.Cardinality;
            }
        }

        #region ctor

        /// <summary>
        /// MUST call from any ctor of inheriting class
        /// </summary>
        private Vector(D dimension)
        {
            Dim = dimension;
            Data = new T[Cardinality];
        }

        /// <summary>
        /// might be useful for vector join node with dynamic input count 
        /// </summary>
        /// <param name="input">A variably typed array of at least the Cardinality's length</param>
        public Vector(D dimension, params T[] input) : this(dimension)
        {
            for (int i = 0; i < Cardinality; i++) Data[i] = input[i];
        }

        /// <summary>
        /// might be useful for vector join node with dynamic input count 
        /// </summary>
        /// <param name="input">A variably typed array of at least the Cardinality's length</param>
        public Vector(D dimension, IEnumerable<T> input) : this(dimension)
        {
            var a = input.Take(Cardinality).ToArray();
            for (int i = 0; i < Cardinality; i++) Data[i] = a[i];
        }


        #endregion

        #region Indexer
        public T  this[int index] {
            get {
                if (index < Cardinality && index >= 0) return Data[index];
                    else throw new ArgumentOutOfRangeException("Vector["+typeof(D).Name+"] has a length of "+Cardinality+". Make sure your Index stays within.");
            }
//          to be truely immutable, this would need to go
            set
            {
                if (index < Cardinality && index >= 0) Data[index] = value;
                    else throw new ArgumentOutOfRangeException("Vector[" + typeof(D).Name + "] has a length of " + Cardinality + ". Make sure your Index stays within.");
            }
        }
        #endregion

        #region Sequence
        
        public IEnumerator<T> GetEnumerator()
        {
            return (IEnumerator<T>)Data.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return Data.GetEnumerator();
        }

        #endregion

        #region utils
        // override in specific implementations, if necessary
        public override string ToString()
        {
            return string.Join(", ", Data.Select(item => item.ToString())).PadLeft(1, '[').PadRight(1, ']');
        }
        #endregion

    }
}
